!---------------------------------------------------------------------!
!                         Containing Element                          !
!                                                                     !
! This module contains procedures to identify which element of a mesh !
! contains a given point.  This version supports meshes in 2D         !
! consisting of triangles and meshes in 3D consisting of tetrahedra.  !
!                                                                     !
! Containing Element is public domain software.  It was produced as   !
! part of work done by the U.S. Government, and is not subject to     !
! copyright in the United States.                                     !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/~WMitchell                                 !
!                                                                     !
!---------------------------------------------------------------------!

module containing_element

implicit none
private
public ce_real, ce_type, ce_create, ce_destroy, ce_add_vertices, &
       ce_add_elements, ce_make_search_data, ce_find_containing_element

!=========================================================================
! Descriptions of public entities
!=========================================================================

! -------
! ce_real is the kind type parameter of real numbers in this module.  All
! real variables passed in or out of the public procedures should be declared
! as real(kind=ce_real) and constants should have this kind parameter
! attached, e.g. 0.0_ce_real

integer, parameter :: ce_real = kind(0.0d0)

! -------
! ce_type is the data structure that contains all the information needed
! by this module.  A variable of this type should be declared and passed
! to the public procedures of this module, e.g. type(ce_type) :: ce

! type ce_type
!    private
!    ...
! end type ce_type

! ---------
! ce_create initializes a containing element data structure.  It must be the
! first procedure called with any variable of type ce_type, and only be called
! once with each such variable until the variable is destroyed with ce_destroy.

! subroutine ce_create(ce, ierr)
! type(ce_type), intent(out) :: ce ! The containing element data structure
! integer, intent(out) :: ierr     ! error code
! Error codes
!  0 no error
!  1 ce was already created
!  2 allocation failed

! ----------
! ce_destroy frees all memory allocated in a containing element data structure.
! It should be called when a variable of type ce_type is no longer needed, and
! only be called once with each such variable until the variable is created
! again with ce_create.

! subroutine ce_destroy(ce, ierr)
! type(ce_type), intent(inout) :: ce ! The containing element data structure
! integer, intent(out) :: ierr       ! error code
! Error codes
!  0 no error
!  1 ce has not been created

! ---------------
! ce_add_vertices adds one or more vertices to the mesh in the containing
! element data structure.  All vertices must be added before calling
! ce_make_search_data.  Vertex IDs must be nonnegative.

! subroutine ce_add_vertices(ce, nvert, verts, ierr, x, y, z)
! type(ce_type), intent(inout) :: ce              ! The containing element data
!                                                 ! structure
! integer, intent(in) :: nvert                    ! The number of vertices given
! integer, intent(in) :: verts(nvert)             ! The vertex ID numbers
! integer, intent(out) :: ierr                    ! error code
! real(ce_real), intent(in) :: x(nvert)           ! x coordinates
! real(ce_real), intent(in), optional :: y(nvert) ! y coordinates if 2D or 3D
! real(ce_real), intent(in), optional :: z(nvert) ! z coordinates if 3D
! Error codes
!  0 no error
!  1 ce has not been created
!  2 nvert is not positive
!  3 one of verts, x, y or z (if present) is dimensioned less than nvert
!  4 a vertex has a different dimension than a previous vertex or element
!  5 a vertex ID is negative
!  6 a vertex ID was previously assigned
!  7 attempt to add vertex after ce_make_search_data called
!  8 allocation failed

! ---------------
! ce_add_elements adds one or more elements to the mesh in the containing
! element data structure.  All elements must be added before calling
! ce_make_search_data.  Element IDs must be nonnegative.  Vertex IDs must
! correspond to a vertex added by ce_add_vertices before or after the
! element is added by ce_add_elements.  The leading dimension of vertices
! must be 3 for triangles or 4 for tetrahedra.

! subroutine ce_add_elements(ce, nelem, elems, vertices, ierr)
! type(ce_type), intent(inout) :: ce                          ! The containing
!                                                             ! element data
!                                                             ! structure
! integer, intent(in) :: nelem                                ! The number of
!                                                             ! elements given
! integer, intent(in) :: elems(nelem)                         ! The element ID
!                                                             ! numbers
! integer, intent(in) :: vertices(vertices per element,nelem) ! The ID numbers
!                                                             ! of the vertices
!                                                             ! of each element
! integer, intent(out) :: ierr                                ! error code
! Error codes
!  0 no error
!  1 ce has not been created
!  2 nelem is not positive
!  3 elems or the second dimension of vertices is less than nelem
!  4 an element has a different dimension than a previous vertex or element, or
!    the first dimension of vertices is not 3 or 4
!  5 an element ID is negative
!  6 an element ID was previously assigned
!  7 attempt to add element after ce_make_search_data called
!  8 allocation failed

! -------------------
! ce_make_search_data builds a data structure that is used when searching for
! an element that contains a point.  It must be called before
! ce_find_containing_element.

! subroutine ce_make_search_data(ce, ierr)
! type(ce_type), intent(inout) :: ce ! The containing element data structure
! integer, intent(out) :: ierr       ! error code
! Error codes
!  0 no error
!  1 ce has not been created
!  2 mismatch in number of elements
!  3 degenerate element found
!  4 allocation failed
!  5 module internal error

! -----------------------
! ce_find_containing_element returns an element that contains the given point,
! or -1 if no element contains it, or a negative valued error code.  If more
! than one element contains the point, which one is returned is arbitrary.

! function ce_find_containing_element(ce, x, y, z)
! type(ce_type), intent(in) :: ce          ! The containing element data
!                                          ! structure
! real(ce_real), intent(in) :: x           ! x coordinate of the point
! real(ce_real), intent(in), optional :: y ! y coordinate of the point if 2D/3D
! real(ce_real), intent(in), optional :: z ! z coordinate of the point if 3D
! integer :: ce_find_containing_element    ! ID number of a containing element
! Negative return values
!  -1 no element contains the point
!  -2 ce has not been created
!  -3 ce_make_search_data has not been called
!  -4 z is present with a 2D mesh, or not present with a 3D mesh
!  -5 degenerate element found
!  -6 module internal error

!=========================================================================
! End of descriptions of public entities
!=========================================================================

! kind number for quad precision real.
! use the second one if quad precision is not available

integer, parameter :: quad_real = selected_real_kind(p=30)
!integer, parameter :: quad_real = selected_real_kind(p=14)

real(ce_real), parameter :: eps = 10*epsilon(0.0_ce_real)

type vertex
   real(ce_real) :: x, y, z ! coordinates
   integer :: nelem         ! number of elements for which this is a vertex
   logical :: in_use        ! whether or not this vertex has been assigned
end type vertex

type element
   integer :: vertex(4) ! the vertices
   logical :: in_use    ! whether or not this element has been assigned
end type element

type bounding_box
   real(ce_real) :: xmin, xmax, ymin, ymax, zmin, zmax
end type bounding_box

type ce_type
   private
   type(vertex), pointer :: vertex(:)       ! the vertices
   type(element), pointer :: element(:)     ! the elements
   type(bounding_box) :: bbox               ! bounding box for the domain
   integer :: dim                           ! 2D or 3D
   integer :: biggest_elem                  ! largest element ID in use
   integer :: nelem                         ! number of elements
   integer :: group_size                    ! desired maximum number of elements
                                            ! in a search group
   integer :: max_tree_depth                ! maximum number of levels in the
                                            ! tree
   integer,pointer :: first_child(:)        ! children for the tree; the root
                                            ! of the tree is node 1; the first
                                            ! child of node i is first_child(i);
                                            ! the other children follow
                                            ! immediately; -1 for leaf
   integer :: nnode                         ! number of nodes in the tree
   integer, pointer :: first_elem(:)        ! first and last element in
   integer, pointer :: last_elem(:)         ! elem_list associated with a node
   integer, pointer :: elem_list(:)         ! that is a leaf
   logical :: created = .false.             ! has create been called?
   logical :: made_search_data              ! has make_search_data been called?
end type ce_type

!----------------------------------------------------

contains

!----------------------------------------------------
! Public routines
!----------------------------------------------------

!          ---------
subroutine ce_create(ce, ierr)
!          ---------

!----------------------------------------------------
! This subroutine initializes a containing element data structure.  It must be
! the first procedure called with any variable of type ce_type, and only be
! called once with each such variable until the variable is destroyed with
! ce_destroy until the variable is created again with ce_create.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(out) :: ce ! The containing element data structure
integer, intent(out) :: ierr     ! error code

!----------------------------------------------------
! Error codes
!  0 no error
!  1 ce was already created
!  2 allocation failed

!----------------------------------------------------
! Local variables

integer :: astat
!----------------------------------------------------
! Begin executable code

ierr = 0

! Check the validity of the input arguments

if (ce%created) then
   ierr = 1
   return
endif

! Start with 100,000 vertices and elements; more will be added if needed.

allocate(ce%vertex(0:100000), ce%element(0:100000), stat=astat)
if (astat /= 0) then
   ierr = 2
   return
endif
ce%vertex%x = 0
ce%vertex%y = 0
ce%vertex%z = 0
ce%vertex%nelem = 0
ce%vertex%in_use = .false.
ce%element%vertex(1) = 0
ce%element%vertex(2) = 0
ce%element%vertex(3) = 0
ce%element%vertex(4) = 0
ce%element%in_use = .false.

! Don't know the dimension until the first vertex is added

ce%dim = -1

! No bounding box yet

ce%bbox%xmin =  huge(0.0_ce_real)
ce%bbox%xmax = -huge(0.0_ce_real)
ce%bbox%ymin =  huge(0.0_ce_real)
ce%bbox%ymax = -huge(0.0_ce_real)
ce%bbox%zmin =  huge(0.0_ce_real)
ce%bbox%zmax = -huge(0.0_ce_real)

! no elements in use

ce%biggest_elem = -1
ce%nelem = 0

! ce_make_search_data has not been called

ce%made_search_data = .false.

! ce_create has been called

ce%created = .true.

end subroutine ce_create

!          ----------
subroutine ce_destroy(ce, ierr)
!          ----------

!----------------------------------------------------
! This subroutine frees all memory allocated in a containing element data
! structure.  It should be called when a variable of type ce_type is no longer
! needed, and only be called once with each such variable.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce ! The containing element data structure
integer, intent(out) :: ierr       ! error code

!----------------------------------------------------
! Error codes
!  0 no error
!  1 ce has not been created

!----------------------------------------------------
! Local variables

!----------------------------------------------------
! Begin executable code

ierr = 0

! Check the validity of the input arguments

if (.not. ce%created) then
   ierr = 1
   return
endif

! Free vertices and elements

deallocate(ce%vertex, ce%element)

! Free the search data

if (ce%made_search_data) then
   deallocate(ce%first_child,ce%first_elem,ce%last_elem,ce%elem_list)
endif

! ce has no longer been created

ce%created = .false.

end subroutine ce_destroy

!          ---------------
subroutine ce_add_vertices(ce, nvert, verts, ierr, x, y, z)
!          ---------------

!----------------------------------------------------
! This subroutine adds one or more vertices to the mesh in the containing
! element data structure.  All vertices must be added before calling
! ce_make_search_data.  Vertex IDs must be nonnegative.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce          ! The containing element data
                                            ! structure
integer, intent(in) :: nvert                ! The number of vertices given
integer, intent(in) :: verts(:)             ! The vertex ID numbers
integer, intent(out) :: ierr                ! error code
real(ce_real), intent(in) :: x(:)           ! x coordinates
real(ce_real), intent(in), optional :: y(:) ! y coordinates if 2D or 3D
real(ce_real), intent(in), optional :: z(:) ! z coordinates if 3D

!----------------------------------------------------
! Error codes
!  0 no error
!  1 ce has not been created
!  2 nvert is not positive
!  3 one of verts, x, y or z (if present) is dimensioned less than nvert
!  4 a vertex has a different dimension than a previous vertex or element
!  5 a vertex ID is negative
!  6 a vertex ID was previously assigned
!  7 attempt to add vertex after ce_make_search_data called
!  8 allocation failed

!----------------------------------------------------
! Local variables

integer :: i, jerr

!----------------------------------------------------
! Begin executable code

! TEMP
if (.not. present(y)) then
   print *,"ERROR: 1D is not yet supported in containing_element"
   stop
endif

ierr = 0

! Check the validity of the input arguments

if (.not. ce%created) then
   ierr = 1
   return
endif

if (nvert < 1) then
   ierr = 2
   return
endif

if (size(verts) < nvert .or. size(x) < nvert .or. size(y) < nvert) then
   ierr = 3
   return
endif

if (present(z)) then
   if (size(z) < nvert) then
      ierr = 3
      return
   endif
endif

if (ce%dim == -1) then
   if (present(z)) then
      ce%dim = 3
   else
      ce%dim = 2
   endif
else
   if ((present(z) .and. ce%dim == 2) .or. &
       (.not. present(z) .and. ce%dim == 3)) then
      ierr = 4
      return
   endif
endif

if (ce%made_search_data) then
   ierr = 7
   return
endif

! Add each vertex to the data structure

do i=1,nvert

! Vertex IDs must be nonnegative

   if (verts(i) < 0) then
      ierr = 5
      return
   endif

! Make sure there is enough space

   if (verts(i) > ubound(ce%vertex,dim=1)) then
      call more_vertices(ce,verts(i),jerr)
      if (jerr /= 0) then
         ierr = 8
         return
      endif
   endif

! Don't reassign a previously assigned vertex

   if (ce%vertex(verts(i))%in_use) then
      ierr = 6
      return
   endif

! Assign the vertex

   ce%vertex(verts(i))%x = x(i)
   ce%vertex(verts(i))%y = y(i)
   if (present(z)) then
      ce%vertex(verts(i))%z = z(i)
   endif
   ce%vertex(verts(i))%in_use = .true.

! Update the bounding box

   ce%bbox%xmin = min(x(i),ce%bbox%xmin)
   ce%bbox%xmax = max(x(i),ce%bbox%xmax)
   ce%bbox%ymin = min(y(i),ce%bbox%ymin)
   ce%bbox%ymax = max(y(i),ce%bbox%ymax)
   if (present(z)) then
      ce%bbox%zmin = min(z(i),ce%bbox%zmin)
      ce%bbox%zmax = max(z(i),ce%bbox%zmax)
   endif

end do

end subroutine ce_add_vertices

!          ---------------
subroutine ce_add_elements(ce, nelem, elems, vertices, ierr)
!          ---------------

!----------------------------------------------------
! This subroutine adds one or more elements to the mesh in the containing
! element data structure.  All elements must be added before calling
! ce_make_search_data.  Element IDs must be nonnegative.  Vertex IDs must
! correspond to a vertex added by ce_add_vertices before or after the
! element is added by ce_add_elements.  The leading dimension of vertices
! must be 3 for triangles or 4 for tetrahedra.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce   ! The containing element data structure
integer, intent(in) :: nelem         ! The number of elements given
integer, intent(in) :: elems(:)      ! The element ID numbers
integer, intent(in) :: vertices(:,:) ! The ID numbers of the vertices of each
                                     ! element
integer, intent(out) :: ierr         ! error code

!----------------------------------------------------
! Error codes
!  0 no error
!  1 ce has not been created
!  2 nelem is not positive
!  3 elems or the second dimension of vertices is less than nelem
!  4 an element has a different dimension than a previous vertex or element, or
!    the first dimension of vertices is not 3 or 4
!  5 an element ID is negative
!  6 an element ID was previously assigned
!  7 attempt to add element after ce_make_search_data called
!  8 allocation failed

!----------------------------------------------------
! Local variables

integer :: i, vert_dim_1, jerr
!----------------------------------------------------
! Begin executable code

ierr = 0

! Check the validity of the input arguments

if (.not. ce%created) then
   ierr = 1
   return
endif

if (nelem < 1) then
   ierr = 2
   return
endif

if (size(elems) < nelem .or. size(vertices,dim=2) < nelem) then
   ierr = 3
   return
endif

vert_dim_1 = size(vertices,dim=1)
if (ce%dim == -1) then
   if (vert_dim_1 /= 3 .and. vert_dim_1 /= 4) then
      ierr = 5
      return
   endif
   ce%dim = vert_dim_1 - 1
else
   if ((ce%dim == 2 .and. vert_dim_1 /= 3) .or. &
       (ce%dim == 3 .and. vert_dim_1 /= 4)) then
      ierr = 4
      return
   endif
endif

if (ce%made_search_data) then
   ierr = 7
   return
endif

! Add each element to the data structure

do i=1,nelem

! Element IDs must be nonnegative

   if (elems(i) < 0) then
      ierr = 5
      return
   endif

! Make sure there is enough space

   if (elems(i) > ubound(ce%element,dim=1)) then
      call more_elements(ce,elems(i),jerr)
      if (jerr /= 0) then
         ierr = 8
         return
      endif
   endif

! Don't reassign a previously assigned element

   if (ce%element(elems(i))%in_use) then
      ierr = 6
      return
   endif

! Assign the element

   ce%element(elems(i))%vertex(1:vert_dim_1) = vertices(:,i)
   ce%element(elems(i))%in_use = .true.

! Increment the number-of-elements count for each vertex, but first make sure
! there is enough space

   if (maxval(vertices(:,i)) > ubound(ce%vertex,dim=1)) then
      call more_vertices(ce,maxval(vertices(:,i)),jerr)
      if (jerr /= 0) then
         ierr = 8
         return
      endif
   endif

   ce%vertex(vertices(:,i))%nelem = ce%vertex(vertices(:,i))%nelem + 1

! Keep track of the biggest element ID seen and number of elements

   ce%biggest_elem = max(elems(i),ce%biggest_elem)
   ce%nelem = ce%nelem + 1

end do

end subroutine ce_add_elements

!          -------------------
subroutine ce_make_search_data(ce, ierr)
!          -------------------

!----------------------------------------------------
! This subroutine builds a data structure that is used when searching for
! an element that contains a point.  It must be called before
! ce_find_containing_element.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce ! The containing element data structure
integer, intent(out) :: ierr       ! error code

!----------------------------------------------------
! Error codes
!  0 no error
!  1 ce has not been created
!  2 mismatch in number of elements
!  3 degenerate element found
!  4 allocation failed
!  5 module internal error

!----------------------------------------------------
! Local variables

integer :: elem_list(ce%nelem)
integer :: nelem, i, next_list, jerr, astat
!----------------------------------------------------
! Begin executable code

ierr = 0

! Check the validity of the input arguments

if (.not. ce%created) then
   ierr = 1
   return
endif

! The desired maximum size of a search group has to be a least as big as
! the maximum number of elements that share a vertex.

ce%group_size = maxval(ce%vertex%nelem) + 2

! The maximum tree depth must be such that the coordinates of two vertices of
! a hexahedron corresponding to a tree node are distinguishable floating point
! numbers.  Each level subdivides each direction into 2 subintervals, so maxlev
! must satisfy 1/2^maxlev > epsilon
! We'll use 10*epsilon

ce%max_tree_depth = log(10*epsilon(0.0_ce_real))/log(0.5_ce_real)

! Start the allocation for the tree nodes at 100000 and increase when needed.

allocate(ce%first_child(100000), ce%first_elem(100000), ce%last_elem(100000), &
         ce%elem_list(100000), stat=astat)
if (astat /= 0) then
   ierr = 4
   return
endif
ce%nnode = 1

! Make a list of all the elements

nelem = 0
do i=1,ce%biggest_elem
   if (ce%element(i)%in_use) then
      nelem = nelem + 1
      elem_list(nelem) = i
   endif
end do
if (nelem /= ce%nelem) then
   ierr = 2
   return
endif

! Make the data

next_list = 1
if (ce%dim == 2) then
   call make_search_data_2D(ce,                       &
                            elem_list, ce%nelem,      & ! all elements
                            next_list,                & ! next starting point
                                                        ! in element list
                            ce%bbox%xmin, ce%bbox%xmax,& ! bounding box
                            ce%bbox%ymin, ce%bbox%ymax,&
                            1,                        & ! first level
                            1,                        & ! root of the tree
                            jerr)                       ! error code
   if (jerr == 1) then
      ierr = 4
      return
   endif
   if (jerr == 2) then
      ierr = 3
      return
   endif
   if (jerr == 3) then
      ierr = 5
      return
   endif
else
   call make_search_data_3D(ce,                       &
                            elem_list, ce%nelem,      & ! all elements
                            next_list,                & ! next starting point
                                                        ! in element list
                            ce%bbox%xmin, ce%bbox%xmax,& ! bounding box
                            ce%bbox%ymin, ce%bbox%ymax,&
                            ce%bbox%zmin, ce%bbox%zmax,&
                            1,                        & ! first level
                            1,                        & ! root of the tree
                            jerr)                       ! error code
   if (jerr == 1) then
      ierr = 4
      return
   endif
   if (jerr == 2) then
      ierr = 3
      return
   endif
   if (jerr == 3) then
      ierr = 5
      return
   endif
endif

! The data has been made

ce%made_search_data = .true.

end subroutine ce_make_search_data

!        --------------------------
function ce_find_containing_element(ce, x, y, z)
!        --------------------------

!----------------------------------------------------
! This function returns an element that contains the given point, or -1 if
! no element contains it, or a negative valued error code.  If more than one
! element contains the point, which one is returned is arbitrary.

!----------------------------------------------------
! Negative return values
!  -1 no element contains the point
!  -2 ce has not been created
!  -3 ce_make_search_data has not been called
!  -4 z is present with a 2D mesh, or not present with a 3D mesh
!  -5 degenerate element found
!  -6 module internal error

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce          ! The containing element data
                                         ! structure
real(ce_real), intent(in) :: x           ! x and y coordinates of the point
real(ce_real), intent(in), optional :: y ! y coordinate of the point if 2D or 3D
real(ce_real), intent(in), optional :: z ! z coordinate of the point if 3D
integer :: ce_find_containing_element    ! ID number of a containing element

!----------------------------------------------------
! Local variables

integer :: jerr, containing_leaf, i, elem

!----------------------------------------------------
! Begin executable code

! TEMP
if (.not. present(y)) then
   print *,"ERROR: 1D not yet supported in containing_element"
   stop
endif

jerr = 0

! Check the validity of the input arguments

if (.not. ce%created) then
   ce_find_containing_element = -2
   return
endif

if (.not. ce%made_search_data) then
   ce_find_containing_element = -3
   return
endif

if (present(z)) then
   if (ce%dim == 2) then
      ce_find_containing_element = -4
      return
   endif
else
   if (ce%dim == 3) then
      ce_find_containing_element = -4
      return
   endif
endif

! If the point is not in the bounding box, then it is not in any element

if (x < ce%bbox%xmin .or. x > ce%bbox%xmax .or. &
    y < ce%bbox%ymin .or. y > ce%bbox%ymax) then
   ce_find_containing_element = -1
   return
endif

if (present(z)) then
   if (z < ce%bbox%zmin .or. z > ce%bbox%zmax) then
      ce_find_containing_element = -1
      return
   endif
endif

! Find the leaf in the octree that contains the point

select case(ce%dim)
case(1)
   containing_leaf = find_containing_leaf(ce,1,x,ce%bbox%xmin,ce%bbox%xmax)
case(2)
   containing_leaf = find_containing_leaf(ce,1,x,ce%bbox%xmin,ce%bbox%xmax, &
                                               y,ce%bbox%ymin,ce%bbox%ymax)
case(3)
   containing_leaf = find_containing_leaf(ce,1,x,ce%bbox%xmin,ce%bbox%xmax, &
                                               y,ce%bbox%ymin,ce%bbox%ymax, &
                                               z,ce%bbox%zmin,ce%bbox%zmax)
case default
   ce_find_containing_element = -6
   return
end select

! See if any of the elements associated with the leaf contain the point, and
! return the index of the first one that does

ce_find_containing_element = -1
if (ce%first_elem(containing_leaf) /= -1) then
   do i=ce%first_elem(containing_leaf),ce%last_elem(containing_leaf)
      elem = ce%elem_list(i)
      if (element_contains_point(ce,elem,x,y,jerr,z)) then
         ce_find_containing_element = elem
         exit
      endif
      if (jerr /= 0) then
         if (jerr == 1) then
            ce_find_containing_element = -5
         else
            ce_find_containing_element = -6
         endif
         exit
      endif
   end do
endif

end function ce_find_containing_element

!----------------------------------------------------
! Private routines
!----------------------------------------------------

!          -------------
subroutine more_vertices(ce, at_least, ierr)
!          -------------

!----------------------------------------------------
! This subroutine increases the allocation for vertices.  Ordinarily it
! doubles the size, but it will make it at least as big as at_least.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout), target :: ce ! The containing element data
                                           ! structure
integer, intent(in) :: at_least            ! Minimum new size
integer, intent(out) :: ierr               ! Nonzero if allocation failed

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed

!----------------------------------------------------
! Local variables

type(vertex), pointer :: old_vertex(:)
integer :: astat, old_size, new_size
!----------------------------------------------------
! Begin executable code

ierr = 0

old_vertex => ce%vertex
old_size = ubound(ce%vertex,dim=1)
new_size = max(2*old_size,at_least)
allocate(ce%vertex(0:new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%vertex(0:old_size) = old_vertex
ce%vertex(old_size+1:new_size)%x = 0
ce%vertex(old_size+1:new_size)%y = 0
ce%vertex(old_size+1:new_size)%z = 0
ce%vertex(old_size+1:new_size)%nelem = 0
ce%vertex(old_size+1:new_size)%in_use = .false.
deallocate(old_vertex)

end subroutine more_vertices

!          -------------
subroutine more_elements(ce, at_least, ierr)
!          -------------

!----------------------------------------------------
! This subroutine increases the allocation for elements.  Ordinarily it
! doubles the size, but it will make it at least as big as at_least.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout), target :: ce ! The containing element data
                                           ! structure
integer, intent(in) :: at_least            ! Minimum new size
integer, intent(out) :: ierr               ! Nonzero if allocation failed

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed

!----------------------------------------------------
! Local variables

type(element), pointer :: old_element(:)
integer :: astat, old_size, new_size
!----------------------------------------------------
! Begin executable code

ierr = 0

old_element => ce%element
old_size = ubound(ce%element,dim=1)
new_size = max(2*old_size,at_least)
allocate(ce%element(0:new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%element(0:old_size) = old_element
ce%element(old_size+1:new_size)%vertex(1) = 0
ce%element(old_size+1:new_size)%vertex(2) = 0
ce%element(old_size+1:new_size)%vertex(3) = 0
ce%element(old_size+1:new_size)%vertex(4) = 0
ce%element(old_size+1:new_size)%in_use = .false.
deallocate(old_element)

end subroutine more_elements

!          ----------
subroutine more_nodes(ce, ierr)
!          ----------

!----------------------------------------------------
! This subroutine increases the allocation for nodes.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout), target :: ce ! The containing element data
                                           ! structure
integer, intent(out) :: ierr               ! Nonzero if allocation failed

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed

!----------------------------------------------------
! Local variables

integer, pointer :: old_nodes(:)
integer, pointer :: old_elem(:)
integer :: astat, old_size, new_size
!----------------------------------------------------
! Begin executable code

ierr = 0

old_nodes => ce%first_child
old_size = ubound(ce%first_child,dim=1)
new_size = 2*old_size
allocate(ce%first_child(new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%first_child(1:old_size) = old_nodes
deallocate(old_nodes)

old_elem => ce%first_elem
allocate(ce%first_elem(new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%first_elem(1:old_size) = old_elem
deallocate(old_elem)

old_elem => ce%last_elem
allocate(ce%last_elem(new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%last_elem(1:old_size) = old_elem
deallocate(old_elem)

end subroutine more_nodes

!          --------------
subroutine more_elem_list(ce, ierr)
!          --------------

!----------------------------------------------------
! This subroutine increases the allocation for elem_list.

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout), target :: ce ! The containing element data
                                           ! structure
integer, intent(out) :: ierr               ! Nonzero if allocation failed

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed

!----------------------------------------------------
! Local variables

integer, pointer :: old_elem(:)
integer :: astat, old_size, new_size
!----------------------------------------------------
! Begin executable code

ierr = 0

old_elem => ce%elem_list
old_size = ubound(ce%elem_list,dim=1)
new_size = 2*old_size
allocate(ce%elem_list(new_size),stat=astat)
if (astat /= 0) then
   ierr = 1
   return
endif
ce%elem_list(1:old_size) = old_elem
deallocate(old_elem)

end subroutine more_elem_list

!                    -------------------
recursive subroutine make_search_data_2D(ce, elem_list, nlist, next_list, &
                                         xmin, xmax, ymin, ymax, level, &
                                         this_node, ierr)
!                    -------------------

!----------------------------------------------------
! This subroutine creates the data that is used by find_containing_element to
! find an element that contains a given point.  In the initial call, elem_list
! is a list of all nlist elements in the mesh, xmin, xmax, ymin and ymax are
! the bounding box, level is 1, and this_node is 1 for the root of the tree.
! In the recursive calls, xmin et al. are the subsquare associated with the
! current tree node, elem_list is a list of all the elements that intersect
! that subsquare, level is the level of this subsquare in the octree, and
! this_node is the new node begin defined.
!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce
integer, intent(in) :: elem_list(:), nlist
integer, intent(inout) :: next_list
real(ce_real), intent(in) :: xmin, xmax, ymin, ymax
integer, intent(in) :: level, this_node
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed
!  2 degenerate element found
!  3 module internal error

!----------------------------------------------------
! Local variables

real(ce_real) :: xlo, xmid, xhi, ylo, ymid, yhi
integer :: new_elem_list(nlist)
integer :: new_nlist, child, jerr, astat

!----------------------------------------------------
! Begin executable code

ierr = 0

! If the number of elements in the list is small enough or the maximum number
! of levels has been reached, then this node is a leaf

if (nlist <= ce%group_size .or. level > ce%max_tree_depth) then

   if (this_node > ubound(ce%first_child,dim=1)) then
      call more_nodes(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif
   ce%first_child(this_node) = -1

! The elements in the list are associated with this tree node

   ce%first_elem(this_node) = next_list
   next_list = next_list + nlist
   ce%last_elem(this_node) = ce%first_elem(this_node) + nlist - 1
   if (ce%last_elem(this_node) > ubound(ce%elem_list,dim=1)) then
      call more_elem_list(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif
   ce%elem_list(ce%first_elem(this_node):ce%last_elem(this_node)) = &
      elem_list(1:nlist)

! If none of the termination criteria are met, then subdivide the square
! and go to each of the children.

else

   xmid = (xmin+xmax)/2
   ymid = (ymin+ymax)/2

   ce%first_child(this_node) = ce%nnode+1
   ce%first_elem(this_node) = -1
   ce%last_elem(this_node) = -1
   ce%nnode = ce%nnode + 4
   if (ce%nnode > ubound(ce%first_child,dim=1)) then
      call more_nodes(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif

   do child=0,3
      select case(child)
      case (0)
         xlo = xmin; xhi = xmid; ylo = ymin; yhi = ymid
      case (1)
         ylo = ymid; yhi = ymax
      case (2)
         xlo = xmid; xhi = xmax
      case (3)
         ylo = ymin; yhi = ymid
      end select
      call make_list_of_intersecting_elements(ce,elem_list,nlist,new_elem_list,&
                                              new_nlist,xlo,xhi,ylo,yhi,jerr)
      if (jerr /= 0) then
         ierr = jerr + 1
         return
      endif
      call make_search_data_2D(ce,new_elem_list,new_nlist,next_list,xlo,xhi, &
                               ylo,yhi,level+1,ce%first_child(this_node)+child,&
                               jerr)
      if (jerr /= 0) then
         ierr = jerr
         return
      endif
   end do

endif

end subroutine make_search_data_2D

!                    -------------------
recursive subroutine make_search_data_3D(ce, elem_list, nlist, next_list, xmin,&
                                         xmax, ymin, ymax, zmin, zmax, level, &
                                         this_node, ierr)
!                    -------------------

!----------------------------------------------------
! This subroutine creates the data that is used by find_containing_element to
! find an element that contains a given point.  In the initial call, elem_list
! is a list of all nlist elements in the mesh, xmin, xmax, ymin, ymax, zmin and
! zmax are the bounding box, level is 1, and this_node is 1 for the root of the
! tree.  In the recursive calls, xmin et al. are the subcube associated with
! the current tree node, elem_list is a list of all the elements that intersect
! that subcube, level is the level of this subsquare in the octree, and
! this_node is the new node begin defined.
!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(inout) :: ce
integer, intent(in) :: elem_list(:), nlist
integer, intent(inout) :: next_list
real(ce_real), intent(in) :: xmin, xmax, ymin, ymax, zmin, zmax
integer, intent(in) :: level, this_node
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 allocation failed
!  2 degenerate element found
!  3 module internal error

!----------------------------------------------------
! Local variables

real(ce_real) :: xlo, xmid, xhi, ylo, ymid, yhi, zlo, zmid, zhi
integer :: new_elem_list(nlist)
integer :: new_nlist, child, jerr, astat

!----------------------------------------------------
! Begin executable code

ierr = 0

! If the number of elements in the list is small enough or the maximum number
! of levels has been reached, then this node is a leaf

if (nlist <= ce%group_size .or. level > ce%max_tree_depth) then

   if (this_node > ubound(ce%first_child,dim=1)) then
      call more_nodes(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif
   ce%first_child(this_node) = -1

! The elements in the list are associated with this tree node

   ce%first_elem(this_node) = next_list
   next_list = next_list + nlist
   ce%last_elem(this_node) = ce%first_elem(this_node) + nlist - 1
   if (ce%last_elem(this_node) > ubound(ce%elem_list,dim=1)) then
      call more_elem_list(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif
   ce%elem_list(ce%first_elem(this_node):ce%last_elem(this_node)) = &
      elem_list(1:nlist)

! If none of the termination criteria are met, then subdivide the square
! and go to each of the children.

else

   xmid = (xmin+xmax)/2
   ymid = (ymin+ymax)/2
   zmid = (zmin+zmax)/2

   ce%first_child(this_node) = ce%nnode+1
   ce%first_elem(this_node) = -1
   ce%last_elem(this_node) = -1
   ce%nnode = ce%nnode + 8
   if (ce%nnode > ubound(ce%first_child,dim=1)) then
      call more_nodes(ce,jerr)
      if (jerr /= 0) then
         ierr = 1
         return
      endif
   endif

   do child=0,7
      select case(child)
      case (0)
         xlo = xmin; xhi = xmid; ylo = ymin; yhi = ymid; zlo = zmin; zhi = zmid
      case (1)
         ylo = ymid; yhi = ymax
      case (2)
         xlo = xmid; xhi = xmax
      case (3)
         ylo = ymin; yhi = ymid
      case (4)
         zlo = zmid; zhi = zmax
      case (5)
         ylo = ymid; yhi = ymax
      case (6)
         xlo = xmin; xhi = xmid
      case (7)
         ylo = ymin; yhi = ymid
      end select
      call make_list_of_intersecting_elements(ce,elem_list,nlist,new_elem_list,&
                                              new_nlist,xlo,xhi,ylo,yhi,jerr, &
                                              zlo,zhi)
      if (jerr /= 0) then
         ierr = jerr + 1
         return
      endif
      call make_search_data_3D(ce,new_elem_list,new_nlist,next_list,xlo,xhi, &
                               ylo,yhi,zlo,zhi,level+1, &
                               ce%first_child(this_node)+child,jerr)
      if (jerr /= 0) then
         ierr = jerr
         return
      endif
   end do

endif

end subroutine make_search_data_3D

!          ----------------------------------
subroutine make_list_of_intersecting_elements(ce, elem_list, nlist, &
                                              new_elem_list, new_nlist, &
                                              xmin, xmax, ymin, ymax, ierr, &
                                              zmin, zmax)
!          ----------------------------------

!----------------------------------------------------
! This routine creates a list of all the elements in elem_list that intersect
! (xmin,xmax) X (ymin,ymax) [ X (zmin,zmax) ].
!
! elem_list is a list of the nlist elements that intersect the parent tree node.
! xmin, xmax, ymin, ymax [zmin, zmax] is one of the four subsquares
! [eight subcubes]  on the current level of the tree.
! new_elem_list and new_nlist are returned as the list of elements that
! intersect the given subsquare [subcube].

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce
integer, intent(in) :: elem_list(:), nlist
integer, intent(out) :: new_elem_list(:), new_nlist
real(ce_real), intent(in) :: xmin, xmax, ymin, ymax
integer, intent(out) :: ierr
real(ce_real), intent(in), optional :: zmin, zmax

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found
!  2 module internal error

!----------------------------------------------------
! Local variables:

integer :: i

!----------------------------------------------------
! Begin executable code

ierr = 0

! Check for mismatch between ce%dim and presence of zmin and zmax

if ((ce%dim == 2 .and. (present(zmin) .or. present(zmax))) .or. &
    (ce%dim == 3 .and. (.not. present(zmin) .or. .not. present(zmax)))) then
   ierr = 2
endif

! For each element in elem_list, determine if it intersects the
! subsquare [subcube], and add it to the new list if it does

new_nlist = 0
do i=1,nlist
   if (ce%dim == 2) then
      if (intersects_triangle(ce,elem_list(i),xmin,xmax,ymin,ymax,ierr)) then
         new_nlist = new_nlist + 1
         new_elem_list(new_nlist) = elem_list(i)
      endif
   else
      if (intersects_tetrahedron(ce,elem_list(i),xmin,xmax,ymin,ymax,zmin, &
                                 zmax,ierr)) then
         new_nlist = new_nlist + 1
         new_elem_list(new_nlist) = elem_list(i)
      endif
   endif
   if (ierr /= 0) exit
end do

end subroutine make_list_of_intersecting_elements

!        -------------------
function intersects_triangle(ce,elem,xmin,xmax,ymin,ymax,ierr)
!        -------------------

!----------------------------------------------------
! This routine determines if triangle element elem intersects the given
! rectangle.

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found
!  2 module internal error

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce
integer, intent(in) :: elem
real(ce_real), intent(in) :: xmin, xmax, ymin, ymax
integer, intent(out) :: ierr
logical :: intersects_triangle

!----------------------------------------------------
! Local variables:

integer :: ivert, side, v1, v2
real(ce_real) :: xminmeps, xmaxmeps, yminmeps, ymaxmeps, &
                 xminpeps, xmaxpeps, yminpeps, ymaxpeps, &
                 xvert(3), yvert(3), xvert1, yvert1, xvert2, yvert2, &
                 min_xvert, max_xvert, min_yvert, max_yvert, &
                 xintersect, yintersect, zeta(3)

!----------------------------------------------------
! Begin executable code

ierr = 0
intersects_triangle = .false.

! "fatten" the rectangle sides slightly for a "fuzzy equals" to avoid missing an
! intersection due to roundoff errors in the input, 
! e.g. x >= xmin becomes x >= xmin-eps and
! x == xmin becomes x >= xmin-eps .and. x <= xmin+eps

xminmeps = xmin - eps
xminpeps = xmin + eps
xmaxmeps = xmax - eps
xmaxpeps = xmax + eps
yminmeps = ymin - eps
yminpeps = ymin + eps
ymaxmeps = ymax - eps
ymaxpeps = ymax + eps

! get the coordinates of the vertices of the triangle

do ivert=1,3
   xvert(ivert) = ce%vertex(ce%element(elem)%vertex(ivert))%x
   yvert(ivert) = ce%vertex(ce%element(elem)%vertex(ivert))%y
end do

! check each vertex of the triangle to see if it is inside the rectangle

do ivert=1,3
   if (xvert(ivert) >= xminmeps .and. xvert(ivert) <= xmaxpeps .and. &
       yvert(ivert) >= yminmeps .and. yvert(ivert) <= ymaxpeps) then
      intersects_triangle = .true.
      exit
   endif
end do

! if a vertex is inside or on the rectangle and one or two of the vertices is
! on a side of the rectangle and the other(s) is outside in the direction of
! that side, don't call it intersecting

if (intersects_triangle) then

   if (abs(xvert(1)-xmin) < eps) then
      if (abs(xvert(2)-xmin) < eps) then
         if (xvert(3) < xmin) then
            intersects_triangle = .false.; return
         endif
      elseif (abs(xvert(3)-xmin) < eps) then
         if (xvert(2) < xmin) then
            intersects_triangle = .false.; return
         endif
      else
         if (xvert(2) < xmin .and. xvert(3) < xmin) then
            intersects_triangle = .false.; return
         endif
      endif
   elseif (abs(xvert(1)-xmax) < eps) then
      if (abs(xvert(2)-xmax) < eps) then
         if (xvert(3) > xmax) then
            intersects_triangle = .false.; return
         endif
      elseif (abs(xvert(3)-xmax) < eps) then
         if (xvert(2) > xmax) then
            intersects_triangle = .false.; return
         endif
      else
         if (xvert(2) > xmax .and. xvert(3) > xmax) then
            intersects_triangle = .false.; return
         endif
      endif
   endif

   if (abs(yvert(1)-ymin) < eps) then
      if (abs(yvert(2)-ymin) < eps) then
         if (yvert(3) < ymin) then
            intersects_triangle = .false.; return
         endif
      elseif (abs(yvert(3)-ymin) < eps) then
         if (yvert(2) < ymin) then
            intersects_triangle = .false.; return
         endif
      else
         if (yvert(2) < ymin .and. yvert(3) < ymin) then
            intersects_triangle = .false.; return
         endif
      endif
   elseif (abs(yvert(1)-ymax) < eps) then
      if (abs(yvert(2)-ymax) < eps) then
         if (yvert(3) > ymax) then
            intersects_triangle = .false.; return
         endif
      elseif (abs(yvert(3)-ymax) < eps) then
         if (yvert(2) > ymax) then
            intersects_triangle = .false.; return
         endif
      else
         if (yvert(2) > ymax .and. yvert(3) > ymax) then
            intersects_triangle = .false.; return
         endif
      endif
   endif

   if (abs(xvert(2)-xmin) < eps) then
      if (abs(xvert(3)-xmin) < eps) then
         if (xvert(1) < xmin) then
            intersects_triangle = .false.; return
         endif
      else
         if (xvert(1) < xmin .and. xvert(3) < xmin) then
            intersects_triangle = .false.; return
         endif
      endif
   elseif (abs(xvert(2)-xmax) < eps) then
      if (abs(xvert(3)-xmax) < eps) then
         if (xvert(1) > xmax) then
            intersects_triangle = .false.; return
         endif
      else
         if (xvert(1) > xmax .and. xvert(3) > xmax) then
            intersects_triangle = .false.; return
         endif
      endif
   endif

   if (abs(yvert(2)-ymin) < eps) then
      if (abs(yvert(3)-ymin) < eps) then
         if (yvert(1) < ymin) then
            intersects_triangle = .false.; return
         endif
      else
         if (yvert(1) < ymin .and. yvert(3) < ymin) then
            intersects_triangle = .false.; return
         endif
      endif
   elseif (abs(yvert(2)-ymax) < eps) then
      if (abs(yvert(3)-ymax) < eps) then
         if (yvert(1) > ymax) then
            intersects_triangle = .false.; return
         endif
      else
         if (yvert(1) > ymax .and. yvert(3) > ymax) then
            intersects_triangle = .false.; return
         endif
      endif
   endif

   if (abs(xvert(3)-xmin) < eps) then
      if (xvert(1) < xmin .and. xvert(2) < xmin) then
         intersects_triangle = .false.; return
      endif
   elseif (abs(xvert(3)-xmax) < eps) then
      if (xvert(1) > xmax .and. xvert(2) > xmax) then
         intersects_triangle = .false.; return
      endif
   endif

   if (abs(yvert(3)-ymin) < eps) then
      if (yvert(1) < ymin .and. yvert(2) < ymin) then
         intersects_triangle = .false.; return
      endif
   elseif (abs(yvert(3)-ymax) < eps) then
      if (yvert(1) > ymax .and. yvert(2) > ymax) then
         intersects_triangle = .false.; return
      endif
   endif

endif

if (intersects_triangle) return

! for each side of the triangle, look for intersections with the sides of
! the rectangle

do side=1,3

! get the coordinates of the endpoints of this side

   v1 = side
   v2 = side+1
   if (v2==4) v2=1
   xvert1 = xvert(v1)
   yvert1 = yvert(v1)
   xvert2 = xvert(v2)
   yvert2 = yvert(v2)
   min_xvert = min(xvert1,xvert2)
   max_xvert = max(xvert1,xvert2)
   min_yvert = min(yvert1,yvert2)
   max_yvert = max(yvert1,yvert2)

! If the line is horizontal, it intersects the left or right if the yverts
! are between ymin and ymax, and max_xvert >= xmin and min_xvert <= xmax.
! And similar for a vertical line.

   if (abs(yvert1-yvert2) < eps) then
      if (yvert1 >= yminmeps .and. yvert1 <= ymaxpeps) then
         if (max_xvert >= xminmeps .and. min_xvert <= xmaxpeps) then
            intersects_triangle = .true.
            exit
         endif
      endif
   endif

   if (abs(xvert1-xvert2) < eps) then
      if (xvert1 >= xminmeps .and. xvert1 <= xmaxpeps) then
         if (max_yvert >= yminmeps .and. min_yvert <= ymaxpeps) then
            intersects_triangle = .true.
            exit
         endif
      endif
   endif

! If max_yvert >= ymin and min_yvert <= ymin then the bottom is a candidate.
! Determine the x value of the point on the line with y value ymin.
! If x value >= xmin and x value <= xmax then it intersects.
! Similar for the other 3 sides of the rectangle.

   if (max_yvert >= yminmeps .and. min_yvert <= yminpeps) then
      xintersect = xvert1 + (ymin-yvert1)*(xvert2-xvert1)/(yvert2-yvert1)
      if (xintersect >= xminmeps .and. xintersect <= xmaxpeps) then
         intersects_triangle = .true.
         exit
      endif
   endif

   if (max_yvert >= ymaxmeps .and. min_yvert <= ymaxpeps) then
      xintersect = xvert1 + (ymax-yvert1)*(xvert2-xvert1)/(yvert2-yvert1)
      if (xintersect >= xminmeps .and. xintersect <= xmaxpeps) then
         intersects_triangle = .true.
         exit
      endif
   endif

   if (max_xvert >= xminmeps .and. min_xvert <= xminpeps) then
      yintersect = yvert1 + (xmin-xvert1)*(yvert2-yvert1)/(xvert2-xvert1)
      if (yintersect >= yminmeps .and. yintersect <= ymaxpeps) then
         intersects_triangle = .true.
         exit
      endif
   endif

   if (max_xvert >= xmaxmeps .and. min_xvert <= xmaxpeps) then
      yintersect = yvert1 + (xmax-xvert1)*(yvert2-yvert1)/(xvert2-xvert1)
      if (yintersect >= yminmeps .and. yintersect <= ymaxpeps) then
         intersects_triangle = .true.
         exit
      endif
   endif

end do

if (intersects_triangle) return

! The only other possibility is that the rectangle is entirely inside the
! triangle.  Check to see if the lower left corner of the rectangle is inside
! by seeing if its barycentric coordinates are all positive.

call barycentric_tri(xmin,ymin,xvert,yvert,zeta,ierr)
if (ierr /= 0) return
if (all(zeta >= -eps)) then
   intersects_triangle = .true.
endif

end function intersects_triangle

!        ----------------------
function intersects_tetrahedron(ce,elem,xmin,xmax,ymin,ymax,zmin,zmax,ierr)
!        ----------------------

!----------------------------------------------------
! This routine determines if tetrahedron element elem intersects the given
! hexahedron.

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found
!  2 module internal error

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce
integer, intent(in) :: elem
real(ce_real), intent(in) :: xmin, xmax, ymin, ymax, zmin, zmax
integer, intent(out) :: ierr
logical :: intersects_tetrahedron

!----------------------------------------------------
! Local variables:

integer :: ivert, iedge, iface
real(ce_real) :: xminmeps, xmaxmeps, yminmeps, ymaxmeps, zminmeps, zmaxmeps, &
                 xminpeps, xmaxpeps, yminpeps, ymaxpeps, zminpeps, zmaxpeps, &
                 xvert(4), yvert(4), zvert(4), xcorner(8), ycorner(8), &
                 zcorner(8), zeta(4), xend(2), yend(2), zend(2)
!----------------------------------------------------
! Begin executable code

ierr = 0
intersects_tetrahedron = .false.

! "fatten" the hexahedron sides slightly for a "fuzzy equals" to avoid missing
! an intersection due to roundoff errors in the input, 
! e.g. x >= xmin becomes x >= xmin-eps and
! x == xmin becomes x >= xmin-eps .and. x <= xmin+eps

xminmeps = xmin - eps
xminpeps = xmin + eps
xmaxmeps = xmax - eps
xmaxpeps = xmax + eps
yminmeps = ymin - eps
yminpeps = ymin + eps
ymaxmeps = ymax - eps
ymaxpeps = ymax + eps
zminmeps = zmin - eps
zminpeps = zmin + eps
zmaxmeps = zmax - eps
zmaxpeps = zmax + eps

! get the coordinates of the vertices of the tetrahedron

do ivert=1,4
   xvert(ivert) = ce%vertex(ce%element(elem)%vertex(ivert))%x
   yvert(ivert) = ce%vertex(ce%element(elem)%vertex(ivert))%y
   zvert(ivert) = ce%vertex(ce%element(elem)%vertex(ivert))%z
end do

! If all the vertices are off to the outside of one of the faces, there is no
! intersection.

if (all(xvert < xminmeps)) return
if (all(xvert > xmaxpeps)) return
if (all(yvert < yminmeps)) return
if (all(yvert > ymaxpeps)) return
if (all(zvert < zminmeps)) return
if (all(zvert > zmaxpeps)) return

! If any vertex of the tetrahedron is inside the hexahedron, it intersects.

do ivert=1,4
   if (xvert(ivert) >= xminmeps .and. xvert(ivert) <= xmaxpeps .and. &
       yvert(ivert) >= yminmeps .and. yvert(ivert) <= ymaxpeps .and. &
       zvert(ivert) >= zminmeps .and. zvert(ivert) <= zmaxpeps) then
      intersects_tetrahedron = .true.
      exit
   endif
end do

if (intersects_tetrahedron) return

! If any vertex of the hexahedron is inside the tetrahedron (i.e. all
! barycentric coordinates are positive), it intersect.

xcorner = (/xmin,xmax,xmax,xmin,xmin,xmax,xmax,xmin/)
ycorner = (/ymin,ymin,ymax,ymax,ymin,ymin,ymax,ymax/)
zcorner = (/zmin,zmin,zmin,zmin,zmax,zmax,zmax,zmax/)
do ivert=1,8
   call barycentric_tet(xcorner(ivert),ycorner(ivert),zcorner(ivert), &
                        xvert,yvert,zvert,zeta,ierr)
   if (ierr /= 0) return
   if (all(zeta >= -eps)) then
      intersects_tetrahedron = .true.
      exit
   endif
end do

if (intersects_tetrahedron) return

! Otherwise, an intersection requires an intersection of an edge of one
! object with a face of the other.

! Check for intersections between edges of the tetrahedron and faces of
! the hexahedron.

outer1: do iface=1,6

! Get the coordinates of the four corners of this face

   select case (iface)
   case (1) ! bottom
      xcorner(1:4) = (/xmin,xmax,xmax,xmin/)
      ycorner(1:4) = (/ymin,ymin,ymax,ymax/)
      zcorner(1:4) = (/zmin,zmin,zmin,zmin/)
   case (2) ! front
      xcorner(1:4) = (/xmin,xmax,xmax,xmin/)
      ycorner(1:4) = (/ymin,ymin,ymin,ymin/)
      zcorner(1:4) = (/zmin,zmin,zmax,zmax/)
   case (3) ! right
      xcorner(1:4) = (/xmax,xmax,xmax,xmax/)
      ycorner(1:4) = (/ymin,ymax,ymax,ymin/)
      zcorner(1:4) = (/zmin,zmin,zmax,zmax/)
   case (4) ! back
      xcorner(1:4) = (/xmin,xmax,xmax,xmin/)
      ycorner(1:4) = (/ymax,ymax,ymax,ymax/)
      zcorner(1:4) = (/zmin,zmin,zmax,zmax/)
   case (5) ! left
      xcorner(1:4) = (/xmin,xmin,xmin,xmin/)
      ycorner(1:4) = (/ymin,ymax,ymax,ymin/)
      zcorner(1:4) = (/zmin,zmin,zmax,zmax/)
   case (6) ! top
      xcorner(1:4) = (/xmin,xmax,xmax,xmin/)
      ycorner(1:4) = (/ymin,ymin,ymax,ymax/)
      zcorner(1:4) = (/zmax,zmax,zmax,zmax/)
   end select

   do iedge=1,6

! Get the coordinates of the endpoints of this edge

      select case(iedge)
      case(1)
         xend = (/xvert(1),xvert(2)/)
         yend = (/yvert(1),yvert(2)/)
         zend = (/zvert(1),zvert(2)/)
      case(2)
         xend = (/xvert(1),xvert(3)/)
         yend = (/yvert(1),yvert(3)/)
         zend = (/zvert(1),zvert(3)/)
      case(3)
         xend = (/xvert(1),xvert(4)/)
         yend = (/yvert(1),yvert(4)/)
         zend = (/zvert(1),zvert(4)/)
      case(4)
         xend = (/xvert(2),xvert(3)/)
         yend = (/yvert(2),yvert(3)/)
         zend = (/zvert(2),zvert(3)/)
      case(5)
         xend = (/xvert(2),xvert(4)/)
         yend = (/yvert(2),yvert(4)/)
         zend = (/zvert(2),zvert(4)/)
      case(6)
         xend = (/xvert(3),xvert(4)/)
         yend = (/yvert(3),yvert(4)/)
         zend = (/zvert(3),zvert(4)/)
      end select

! Check for intersection

      if (line_intersects_rectangle(xend,yend,zend,xcorner(1:4), &
                                    ycorner(1:4),zcorner(1:4),ierr)) then
         intersects_tetrahedron = .true.
         exit outer1
      endif

      if (ierr == 1) then
         ierr = 2
         return
      endif

   end do ! iedge
end do outer1 ! iface

if (intersects_tetrahedron) return

! Check for intersections between edges of the hexahedron and faces of
! the tetrahedron.

outer2: do iedge=1,12

! Get the endpoints of the edges.  Go around the bottom, then around the top,
! then the vertical edges.

   select case (iedge)
   case (1)
      xend = (/xmin,xmax/); yend = (/ymin,ymin/); zend = (/zmin,zmin/)
   case (2)
      xend = (/xmax,xmax/); yend = (/ymin,ymax/); zend = (/zmin,zmin/)
   case (3)
      xend = (/xmin,xmax/); yend = (/ymax,ymax/); zend = (/zmin,zmin/)
   case (4)
      xend = (/xmin,xmin/); yend = (/ymin,ymax/); zend = (/zmin,zmin/)
   case (5)
      xend = (/xmin,xmax/); yend = (/ymin,ymin/); zend = (/zmax,zmax/)
   case (6)
      xend = (/xmax,xmax/); yend = (/ymin,ymax/); zend = (/zmax,zmax/)
   case (7)
      xend = (/xmin,xmax/); yend = (/ymax,ymax/); zend = (/zmax,zmax/)
   case (8)
      xend = (/xmin,xmin/); yend = (/ymin,ymax/); zend = (/zmax,zmax/)
   case (9)
      xend = (/xmin,xmin/); yend = (/ymin,ymin/); zend = (/zmin,zmax/)
   case (10)
      xend = (/xmax,xmax/); yend = (/ymin,ymin/); zend = (/zmin,zmax/)
   case (11)
      xend = (/xmax,xmax/); yend = (/ymax,ymax/); zend = (/zmin,zmax/)
   case (12)
      xend = (/xmin,xmin/); yend = (/ymax,ymax/); zend = (/zmin,zmax/)
   end select

   do iface=1,4

! Get the coordinates of the three corners of this face.

      select case(iface)
      case (1)
         xcorner(1:3) = (/xvert(1),xvert(2),xvert(3)/)
         ycorner(1:3) = (/yvert(1),yvert(2),yvert(3)/)
         zcorner(1:3) = (/zvert(1),zvert(2),zvert(3)/)
      case (2)
         xcorner(1:3) = (/xvert(1),xvert(2),xvert(4)/)
         ycorner(1:3) = (/yvert(1),yvert(2),yvert(4)/)
         zcorner(1:3) = (/zvert(1),zvert(2),zvert(4)/)
      case (3)
         xcorner(1:3) = (/xvert(1),xvert(3),xvert(4)/)
         ycorner(1:3) = (/yvert(1),yvert(3),yvert(4)/)
         zcorner(1:3) = (/zvert(1),zvert(3),zvert(4)/)
      case (4)
         xcorner(1:3) = (/xvert(2),xvert(3),xvert(4)/)
         ycorner(1:3) = (/yvert(2),yvert(3),yvert(4)/)
         zcorner(1:3) = (/zvert(2),zvert(3),zvert(4)/)
      end select

! Check for intersection

      if (line_intersects_triangle(xend,yend,zend,xcorner(1:3), &
                                   ycorner(1:3),zcorner(1:3),ierr)) then
         intersects_tetrahedron = .true.
         exit outer2
      endif
      if (ierr /= 0) return

   end do ! iface
end do outer2 ! iedge

end function intersects_tetrahedron

!        -------------------------
function line_intersects_rectangle(xend,yend,zend,xcorner,ycorner,zcorner,ierr)
!        -------------------------

! Determine if the line with endpoints (xend,yend,zend) intersects the
! rectangle with corners (xcorner,ycorner,zcorner).  The rectangle is
! parallel to one of the coordinate planes, i.e. one of xcorner, ycorner or
! zcorner must be a constant, and has edges parallel to the axes.  The order
! of the corners must be such that corner(1) = (min,min), corner(3) = (max,max)

!----------------------------------------------------
! Error codes
!  0 no error
!  1 module internal error

real(ce_real), intent(in) :: xend(2), yend(2), zend(2), xcorner(4), &
                             ycorner(4), zcorner(4)
integer, intent(out) :: ierr
logical :: line_intersects_rectangle

ierr = 0
line_intersects_rectangle = .false.

! Determine which coordinate is constant and check for intersection

if (all(xcorner == xcorner(1))) then
   line_intersects_rectangle = reduced_line_intersects_rectangle( &
                               yend,zend,xend,ycorner(1),ycorner(3), &
                               zcorner(1),zcorner(3),xcorner(1))
elseif (all(ycorner == ycorner(1))) then
   line_intersects_rectangle = reduced_line_intersects_rectangle( &
                               xend,zend,yend,xcorner(1),xcorner(3), &
                               zcorner(1),zcorner(3),ycorner(1))
elseif (all(zcorner == zcorner(1))) then
   line_intersects_rectangle = reduced_line_intersects_rectangle( &
                               xend,yend,zend,xcorner(1),xcorner(3), &
                               ycorner(1),ycorner(3),zcorner(1))
else

! Don't have a constant coordinate

   ierr = 1
   return
   
endif

end function line_intersects_rectangle

!        ---------------------------------
function reduced_line_intersects_rectangle(uend,vend,wend,umin,umax,vmin,vmax, &
                                           wconst)
!        ---------------------------------

! Determine if the line with endpoints (uend,vend,wend) intersects the
! rectangle (umin,vmin,wconst)X(umax,vmax,wconst)

real(ce_real), intent(in) :: uend(2), vend(2), wend(2), umin, umax, vmin, &
                             vmax, wconst
logical :: reduced_line_intersects_rectangle

real(ce_real) :: a, b, c, u, v, value
integer :: i
logical :: positive

! Can't intersect if both endpoints are above or below wconst

if ((wend(1) > wconst+eps .and. wend(2) > wconst+eps) .or. &
    (wend(1) < wconst-eps .and. wend(2) < wconst-eps)) then
   reduced_line_intersects_rectangle = .false.
else

! If the line lies in the wconst plane, then 1) there is no intersection if
! both endpoints are to one side of the rectangle, 2) there is an intersection
! if corners of the rectangle lie on different sides of the line, i.e. have
! different signs when plugged into the equation of line a*u + b*v + c = 0,
! or if any one of them lies on the line.
! Note that if there are corners on opposite sides of the infinite line and
! the line segment does not intersect the rectangle, then both endpoints must
! be to one side of the rectangle.

   if (abs(wend(1)-wconst) <= eps .and. abs(wend(2)-wconst) <= eps) then

      if ((uend(1) < umin-eps .and. uend(2) < umin-eps) .or. &
          (uend(1) > umax+eps .and. uend(2) > umax+eps) .or. &
          (vend(1) < vmin-eps .and. vend(2) < vmin-eps) .or. &
          (vend(1) > vmax+eps .and. vend(2) > vmax+eps)) then
         reduced_line_intersects_rectangle = .false.

      else

         a = vend(2) - vend(1)
         b = uend(1) - uend(2)
         c = -(a*uend(1) + b*vend(1))

         reduced_line_intersects_rectangle = .false.
         do i=1,4
            select case (i)
            case (1)
               u=umin; v=vmin
            case (2)
               u=umax; v=vmin
            case (3)
               u=umin; v=vmax
            case (4)
               u=umax; v=vmax
            end select
            value = a*u + b*v + c
            if (abs(value) < eps) then
               reduced_line_intersects_rectangle = .true.
               exit
            endif
            if (i==1) then
               positive = value > -eps
            else
               if ((positive .and. value < eps) .or. &
                   (.not. positive .and. value > -eps)) then
                  reduced_line_intersects_rectangle = .true.
                  exit
               endif
            endif
         end do

      endif ! both endpoints lie to one side

! If the line does not lie in the wconst plane, find the (u,v) where it
! intersects the plane and see if that is in the rectangle.

   else

      u = (wconst-wend(1))*(uend(2)-uend(1))/(wend(2)-wend(1)) + uend(1)
      if (u < umin-eps .or. u > umax+eps) then
         reduced_line_intersects_rectangle = .false.
      else
         v = (wconst-wend(1))*(vend(2)-vend(1))/(wend(2)-wend(1)) + vend(1)
         if (v < vmin-eps .or. v > vmax+eps) then
            reduced_line_intersects_rectangle = .false.
         else
            reduced_line_intersects_rectangle = .true.
         endif
      endif

   endif

endif

end function reduced_line_intersects_rectangle

!        ------------------------
function line_intersects_triangle(xend,yend,zend,xvert,yvert,zvert,ierr)
!        ------------------------

! Determine if the line with endpoints (xend,yend,zend) intersects the
! triangle with vertices (xvert,yvert,zvert).  The line is parallel to one
! of the axes, i.e., only one of xend, yend and zend can have end(1) /= end(2).

real(ce_real), intent(in) :: xend(2),yend(2),zend(2),xvert(3),yvert(3),zvert(3)
integer, intent(out) :: ierr
logical :: line_intersects_triangle

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found
!  2 module internal error

ierr = 0
line_intersects_triangle = .false.

! Determine which axis the line is parallel to and check for intersection

if (xend(1) /= xend(2)) then
   if (yend(1) /= yend(2) .or. zend(1) /= zend(2)) then
      ierr = 2 ! line not parallel to an axis"
      return
   endif
   line_intersects_triangle =  reduced_line_intersects_triangle( &
                               yend(1),zend(1),xend,yvert,zvert,xvert,ierr)
   if (ierr /= 0) return
elseif (yend(1) /= yend(2)) then
   if (xend(1) /= xend(2) .or. zend(1) /= zend(2)) then
      ierr = 2 ! line not parallel to an axis"
      return
   endif
   line_intersects_triangle =  reduced_line_intersects_triangle( &
                               xend(1),zend(1),yend,xvert,zvert,yvert,ierr)
   if (ierr /= 0) return
elseif (zend(1) /= zend(2)) then
   if (xend(1) /= xend(2) .or. yend(1) /= yend(2)) then
      ierr = 2 ! line not parallel to an axis"
      return
   endif
   line_intersects_triangle =  reduced_line_intersects_triangle( &
                               xend(1),yend(1),zend,xvert,yvert,zvert,ierr)
   if (ierr /= 0) return
else
   ierr = 2 ! line not parallel to an axis"
   return
endif

end function line_intersects_triangle

!        --------------------------------
function reduced_line_intersects_triangle(uend,vend,wend,uvert,vvert,wvert,ierr)
!        --------------------------------

! Determine if the vertical line (uend,vend,wend(1)),(uend,vend,wend(2))
! intersects the triangle with vertices (uvert,vvert,wvert)

real(ce_real), intent(in) :: uend, vend, wend(2), uvert(3), vvert(3), wvert(3)
integer, intent(out) :: ierr
logical :: reduced_line_intersects_triangle

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found
!  2 module internal error

real(ce_real) :: zeta(3), AB(3), AC(3), a, b, c, d, x1, x2, y1, y2
integer :: i

ierr = 0
reduced_line_intersects_triangle = .false.

! wend(1) must be less than wend(2)

! line must point upwards

if (wend(1) >= wend(2)) then
   ierr = 2
   return
endif

! If all vertices are below or above the line, can't intersect

if (all(wvert < wend(1)-eps) .or. all(wvert > wend(2)+eps)) then
   reduced_line_intersects_triangle = .false.
else

! The projection onto the u-v plane must have the point (uend,vend) inside
! the triangle with vertices (uvert,vvert,0)

   call barycentric_tri(uend,vend,uvert,vvert,zeta,ierr)
   if (ierr == 0) then
      if (all(zeta(:) >= -eps)) then

! Determine if the line segment goes through the plane of the triangle

! The equation of the plane a*u + b*v + c*w + d = 0

         AB(1) = uvert(2)-uvert(1)
         AB(2) = vvert(2)-vvert(1)
         AB(3) = wvert(2)-wvert(1)
         AC(1) = uvert(3)-uvert(1)
         AC(2) = vvert(3)-vvert(1)
         AC(3) = wvert(3)-wvert(1)
         a = AB(2)*AC(3) - AC(2)*AB(3)
         b = -(AB(1)*AC(3) - AC(1)*AB(3))
         c = AB(1)*AC(2) - AC(1)*AB(2)
         d = -(a*uvert(1)+b*vvert(1)+c*wvert(1))

! Intersect if the result of entering the endpoints into the equation give
! different signs or 0, i.e. the product is negative or 0.

         reduced_line_intersects_triangle = (a*uend+b*vend+c*wend(1)+d) * &
                                            (a*uend+b*vend+c*wend(2)+d) < eps

      else ! projection on u-v plane does not have the point in the triangle

         reduced_line_intersects_triangle = .false.

      endif

   else

! If barycentric_tri returned an error code, then the triangle (uvert,vvert,0)
! is degenerate, which means the triangle is perpendicular to the plane.  In
! that case the point (uend,vend) must lie on the line given by uvert and vvert,
! i.e. it must be a linear combination of two of the points with coefficients
! between 0 and 1.

      ierr = 0
      do i=1,3
         select case(i)
         case(1)
            x1 = uvert(1)
            y1 = vvert(1)
            x2 = uvert(2)
            y2 = vvert(2)
         case(2)
            x1 = uvert(1)
            y1 = vvert(1)
            x2 = uvert(3)
            y2 = vvert(3)
         case(3)
            x1 = uvert(2)
            y1 = vvert(2)
            x2 = uvert(3)
            y2 = vvert(3)
         end select
         if (abs(x1-x2) < eps) then
            if (abs(uend-x2) < eps) then
               if (abs(y1-y2) > eps) then
                  a = (vend-y2)/(y1-y2)
                  if (a > -eps .and. a < 1.0+eps) then
                     reduced_line_intersects_triangle = .true.
                     exit
                  endif
               endif
            endif
         else
            a = (uend-x2)/(x1-x2)
            if (a > -eps .and. a < 1.0+eps) then
               if (abs(vend-a*y1-(1-a)*y2) < eps) then
                  reduced_line_intersects_triangle = .true.
                  exit
               endif
            endif
         endif
      end do
   endif
endif

end function reduced_line_intersects_triangle

!                  --------------------
recursive function find_containing_leaf(ce,node,x,xmin,xmax,y,ymin,ymax, &
                                        z,zmin,zmax) result(res)
!                  --------------------

!----------------------------------------------------
! This routine descends the tree to find the leaf that contains the given point.
! node is the current node of the tree.  xmin, xmax, etc. are the extents of
! the box associated with this node.
!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce
real(ce_real), intent(in) :: x, xmin, xmax
real(ce_real), intent(in), optional :: y, ymin, ymax, z, zmin, zmax
integer, intent(in) :: node
integer :: res

!----------------------------------------------------
! Local variables:

real(ce_real) :: xmid, ymid, zmid, cxmin, cxmax, cymin, cymax, czmin, czmax
integer :: code, child
integer :: order(0:7) = (/0,3,1,2,7,4,6,5/)
!----------------------------------------------------
! Begin executable code

! if this is a leaf, we have found it

if (ce%first_child(node) == -1) then
   res = node
   return
endif

! midpoints of the extents

xmid = (xmin+xmax)/2
if (present(y)) ymid = (ymin+ymax)/2
if (present(z)) zmid = (zmin+zmax)/2

! code indicates which subdivision of node contains the point.  The lowest order
! bit is 0 if x < xmid and 1 if x >= xmid.  The second bit is the same with y,
! etc.  Also set the extents of the subdivision.

if (x < xmid) then
   code = 0
   cxmin = xmin; cxmax = xmid
else
   code = 1
   cxmin = xmid; cxmax = xmax
endif
if (present(y)) then
   if (y < ymid) then
      cymin = ymin; cymax = ymid
   else
      code = code + 2
      cymin = ymid; cymax = ymax
   endif
endif
if (present(z)) then
   if (z < zmid) then
      czmin = zmin; czmax = zmid
   else
      code = code + 4
      czmin = zmid; czmax = zmax
   endif
endif

! order gives the local child number (0..7) for the subdivision indicated by
! code.  order is the same order used in make_search_data_xD.
! child is then the node of the tree that contains the given point.

child = ce%first_child(node) + order(code)

! descend to the child

select case(ce%dim)
case(1)
   res = find_containing_leaf(ce,child,x,cxmin,cxmax)
case(2)
   res = find_containing_leaf(ce,child,x,cxmin,cxmax,y,cymin,cymax)
case(3)
   res = find_containing_leaf(ce,child,x,cxmin,cxmax,y,cymin,cymax,z,czmin,czmax)
end select

end function find_containing_leaf

!        ----------------------
function element_contains_point(ce,elem,x,y,ierr,z)
!        ----------------------

!----------------------------------------------------
! This routine determines whether or not element elem contains the given point

!----------------------------------------------------
! Dummy arguments

type(ce_type), intent(in) :: ce
integer, intent(in) :: elem
real(ce_real), intent(in) :: x, y
real(ce_real), intent(in), optional :: z
integer, intent(out) :: ierr
logical :: element_contains_point

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found

!----------------------------------------------------
! Local variables:

real(ce_real) :: zeta(ce%dim+1), zz
!----------------------------------------------------
! Begin executable code

ierr = 0

if (x < minval(ce%vertex(ce%element(elem)%vertex)%x)-eps) then
   element_contains_point = .false.
   return
endif
if (x > maxval(ce%vertex(ce%element(elem)%vertex)%x)+eps) then
   element_contains_point = .false.
   return
endif
if (y < minval(ce%vertex(ce%element(elem)%vertex)%y)-eps) then
   element_contains_point = .false.
   return
endif
if (y > maxval(ce%vertex(ce%element(elem)%vertex)%y)+eps) then
   element_contains_point = .false.
   return
endif
if (present(z)) then
   if (z < minval(ce%vertex(ce%element(elem)%vertex)%z)-eps) then
      element_contains_point = .false.
      return
   endif
   if (z > maxval(ce%vertex(ce%element(elem)%vertex)%z)+eps) then
      element_contains_point = .false.
      return
   endif
endif

! Compute the barycentric coordinates of the given point with respect to
! the element vertices

if (present(z)) then
   zz = z ! because z is not optional in barycentric_tet
   call barycentric_tet(x,y,zz,ce%vertex(ce%element(elem)%vertex)%x, &
                               ce%vertex(ce%element(elem)%vertex)%y, &
                               ce%vertex(ce%element(elem)%vertex)%z,zeta,ierr)
   if (ierr /= 0) return
else
   call barycentric_tri(x,y,ce%vertex(ce%element(elem)%vertex)%x, &
                            ce%vertex(ce%element(elem)%vertex)%y,zeta,ierr)
   if (ierr /= 0) return
endif

! The element contains the point if all barycentric coordinates are positive.
! Allow a slight margin for roundoff error.

element_contains_point = all(zeta >= -eps)

end function element_contains_point

!          ---------------
subroutine barycentric_tri(x,y,xvert,yvert,zeta,ierr)
!          ---------------

!----------------------------------------------------
! This subroutine returns the barycentric coordinates of (x,y) in the
! triangle with the given vertices

!----------------------------------------------------
! Dummy arguments

real(ce_real), intent(in) :: x,y,xvert(3),yvert(3)
real(ce_real), intent(out) :: zeta(3)
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found

!----------------------------------------------------
! Local variables:

real(ce_real) :: x1,x2,x3,y1,y2,y3,det
real(ce_real), parameter :: singleeps = epsilon(0.0)

!----------------------------------------------------
! Begin executable code

ierr = 0

! If the triangle is smallish, use the more accurate version.

if (maxval(xvert)-minval(xvert) < singleeps .or. &
    maxval(yvert)-minval(yvert) < singleeps) then
   call accurate_barycentric_tri(x,y,xvert,yvert,zeta,ierr)
   return
endif

! put the vertices in local variables to shorten expressions

x1 = xvert(1)
x2 = xvert(2)
x3 = xvert(3)
y1 = yvert(1)
y2 = yvert(2)
y3 = yvert(3)

! compute the barycentric coordinates of the point

! compute the determinant

det = x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2)

if (det == 0.0_ce_real) then
   ierr = 1
   return
endif

! compute the coordinates

zeta(1) = (x*(y2-y3) + y*(x3-x2) + (x2*y3-x3*y2))/det
zeta(2) = (x*(y3-y1) + y*(x1-x3) + (x3*y1-x1*y3))/det
zeta(3) = 1.0_ce_real - zeta(1) - zeta(2)

! if all the zeta are sort of close to being positive, but not real close
! to being positive, do a more accurate computation

if (.not. all(zeta >= -eps)) then
   if (all(zeta > -1.0e-6_ce_real)) then
      call accurate_barycentric_tri(x,y,xvert,yvert,zeta,ierr)
   endif
endif

end subroutine barycentric_tri

!          ------------------------
subroutine accurate_barycentric_tri(x,y,xvert,yvert,zeta,ierr)
!          ------------------------

!----------------------------------------------------
! This subroutine returns the barycentric coordinates of (x,y)
! in the triangle with vertices (xvert,yvert).
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

real(ce_real), intent(in) :: x, y, xvert(3), yvert(3)
real(ce_real), intent(out) :: zeta(3)
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found

!----------------------------------------------------
! Local variables:

real(quad_real) :: x1,x2,x3,y1,y2,y3,det,ssump,ssumm
real(quad_real) :: xy1,xy2,xy3,yx1,yx2,yx3,sump,summ
real(quad_real) :: x1y2,x2y3,x3y1,x1y3,x2y1,x3y2
real(quad_real) :: zeta1, zeta2

!----------------------------------------------------
! Begin executable code

ierr = 0

! put the vertices in local variables to shorten expressions

x1 = real(xvert(1),quad_real)
x2 = real(xvert(2),quad_real)
x3 = real(xvert(3),quad_real)
y1 = real(yvert(1),quad_real)
y2 = real(yvert(2),quad_real)
y3 = real(yvert(3),quad_real)

! compute the barycentric coordinates of the point

! reduce roundoff by summing all the positive parts and negative parts
! separately, and then adding the two partial sums

! products needed for the sums

xy1 = real(x,quad_real)*y1
xy2 = real(x,quad_real)*y2
xy3 = real(x,quad_real)*y3
yx1 = real(y,quad_real)*x1
yx2 = real(y,quad_real)*x2
yx3 = real(y,quad_real)*x3
x1y2 = x1*y2
x2y3 = x2*y3
x3y1 = x3*y1
x1y3 = x1*y3
x2y1 = x2*y1
x3y2 = x3*y2

! compute the determinant

! det = x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2)

ssump=0.0_quad_real; ssumm=0.0_quad_real
if (x1y2 > 0.0_quad_real) then; ssump=ssump+x1y2; else; ssumm=ssumm+x1y2; endif
if (x1y3 < 0.0_quad_real) then; ssump=ssump-x1y3; else; ssumm=ssumm-x1y3; endif
if (x2y3 > 0.0_quad_real) then; ssump=ssump+x2y3; else; ssumm=ssumm+x2y3; endif
if (x2y1 < 0.0_quad_real) then; ssump=ssump-x2y1; else; ssumm=ssumm-x2y1; endif
if (x3y1 > 0.0_quad_real) then; ssump=ssump+x3y1; else; ssumm=ssumm+x3y1; endif
if (x3y2 < 0.0_quad_real) then; ssump=ssump-x3y2; else; ssumm=ssumm-x3y2; endif
det = ssump + ssumm

if (real(det,ce_real) == 0.0_ce_real) then
   ierr = 1
   return
endif

! compute the coordinates

! zeta(1) = (x*(y2-y3) + y*(x3-x2) + (x2*y3-x3*y2))/det

sump=0.0_quad_real; summ=0.0_quad_real
if (xy2 > 0.0_quad_real) then
   sump=sump+xy2
else
   summ=summ+xy2
endif
if (xy3 < 0.0_quad_real) then
   sump=sump-xy3
else
   summ=summ-xy3
endif
if (yx3 > 0.0_quad_real) then
   sump=sump+yx3
else
   summ=summ+yx3
endif
if (yx2 < 0.0_quad_real) then
   sump=sump-yx2
else
   summ=summ-yx2
endif
if (x2y3 > 0.0_quad_real) then; sump=sump+x2y3; else; summ=summ+x2y3; endif
if (x3y2 < 0.0_quad_real) then; sump=sump-x3y2; else; summ=summ-x3y2; endif
zeta1 = (sump + summ)/det
zeta(1) = zeta1

! zeta(2) = (x*(y3-y1) + y*(x1-x3) + (x3*y1-x1*y3))/det

sump=0.0_quad_real; summ=0.0_quad_real
if (xy3 > 0.0_quad_real) then
   sump=sump+xy3
else
   summ=summ+xy3
endif
if (xy1 < 0.0_quad_real) then
   sump=sump-xy1
else
   summ=summ-xy1
endif
if (yx1 > 0.0_quad_real) then
   sump=sump+yx1
else
   summ=summ+yx1
endif
if (yx3 < 0.0_quad_real) then
   sump=sump-yx3
else
   summ=summ-yx3
endif
if (x3y1 > 0.0_quad_real) then; sump=sump+x3y1; else; summ=summ+x3y1; endif
if (x1y3 < 0.0_quad_real) then; sump=sump-x1y3; else; summ=summ-x1y3; endif
zeta2 = (sump + summ)/det
zeta(2) = zeta2

! zeta(3) = (x*(y1-y2) + y*(x2-x1) + (x1*y2-x2*y1))/det
! but easier is zeta(3) = 1 - zeta(1) - zeta(2)

!sump=0.0_quad_real; summ=0.0_quad_real
!if (xy1 > 0.0_quad_real) then
!   sump=sump+xy1
!else
!   summ=summ+xy1
!endif
!if (xy2 < 0.0_quad_real) then
!   sump=sump-xy2
!else
!   summ=summ-xy2
!endif
!if (yx2 > 0.0_quad_real) then
!   sump=sump+yx2
!else
!   summ=summ+yx2
!endif
!if (yx1 < 0.0_quad_real) then
!   sump=sump-yx1
!else
!   summ=summ-yx1
!endif
!if (x1y2 > 0.0_quad_real) then; sump=sump+x1y2; else; summ=summ+x1y2; endif
!if (x2y1 < 0.0_quad_real) then; sump=sump-x2y1; else; summ=summ-x2y1; endif
!zeta(3) = sump/det + summ/det

zeta(3) = 1.0_quad_real - zeta(1) - zeta(2)

end subroutine accurate_barycentric_tri

!          ---------------
subroutine barycentric_tet(x,y,z,xv,yv,zv,zeta,ierr)
!          ---------------

!----------------------------------------------------
! This function returns the barycentric coordinates of (x,y,z) in the
! tetrahedral element with vertices (xv,yv,zv).
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

real(ce_real), intent(in) :: x,y,z,xv(4),yv(4),zv(4)
real(ce_real), intent(out) :: zeta(4)
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found

!----------------------------------------------------
! Local variables:

real(ce_real) :: A11, A12, A13, A21, A22, A23, A31, A32, A33, b1, b2, b3, det
integer :: ipiv(3), info, i
!----------------------------------------------------
! Begin executable code

ierr = 0

! set up the linear system whose solution determines the first 3 coordinates

A11 = xv(1) - xv(4); A12 = xv(2) - xv(4); A13 = xv(3) - xv(4)
A21 = yv(1) - yv(4); A22 = yv(2) - yv(4); A23 = yv(3) - yv(4)
A31 = zv(1) - zv(4); A32 = zv(2) - zv(4); A33 = zv(3) - zv(4)

b1 = x - xv(4)
b2 = y - yv(4)
b3 = z - zv(4)

! compute the barycentric coordinates of the point

det = A11*(A22*A33-A32*A23) - &
      A21*(A12*A33-A32*A13) + &
      A31*(A12*A23-A22*A13)

if (det == 0.0_ce_real) then
   ierr = 1
   return
endif

! compute the coordinates

zeta(1) = (b1*(A22*A33-A32*A23) - &
           b2*(A12*A33-A32*A13) + &
           b3*(A12*A23-A22*A13))/det

zeta(2) = (A11*(b2*A33-b3*A23) - &
           A21*(b1*A33-b3*A13) + &
           A31*(b1*A23-b2*A13))/det

zeta(3) = (A11*(A22*b3-A32*b2) - &
           A21*(A12*b3-A32*b1) + &
           A31*(A12*b2-A22*b1))/det

zeta(4) = 1.0_ce_real - zeta(1) - zeta(2) - zeta(3)

! if all the zeta are sort of close to being positive, but not real close
! to being positive, do a more accurate computation

if (.not. all(zeta >= -eps)) then
   if (all(zeta > -1.0e-6_ce_real)) then
      call accurate_barycentric_tet(x,y,z,xv,yv,zv,zeta,ierr)
   endif
endif

end subroutine barycentric_tet

!          ------------------------
subroutine accurate_barycentric_tet(x,y,z,xv,yv,zv,zeta,ierr)
!          ------------------------

!----------------------------------------------------
! This function returns the barycentric coordinates of (x,y,z) in the
! tetrahedral element with vertices (xv,yv,zv).
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

real(ce_real), intent(in) :: x,y,z,xv(4),yv(4),zv(4)
real(ce_real), intent(out) :: zeta(4)
integer, intent(out) :: ierr

!----------------------------------------------------
! Error codes
!  0 no error
!  1 degenerate element found

!----------------------------------------------------
! Local variables:

real(quad_real) :: A11, A12, A13, A21, A22, A23, A31, A32, A33, b1, b2, b3, det
real(quad_real) :: xv4, yv4, zv4, temp, sump, summ, zeta1, zeta2, zeta3
integer :: ipiv(3), info, i
!----------------------------------------------------
! Begin executable code

ierr = 0

! set up the linear system whose solution determines the first 3 coordinates

xv4 = real(xv(4),quad_real)
yv4 = real(yv(4),quad_real)
zv4 = real(zv(4),quad_real)

A11 = real(xv(1),quad_real) - xv4
A12 = real(xv(2),quad_real) - xv4
A13 = real(xv(3),quad_real) - xv4
A21 = real(yv(1),quad_real) - yv4
A22 = real(yv(2),quad_real) - yv4
A23 = real(yv(3),quad_real) - yv4
A31 = real(zv(1),quad_real) - zv4
A32 = real(zv(2),quad_real) - zv4
A33 = real(zv(3),quad_real) - zv4

b1 = real(x,quad_real) - xv4
b2 = real(y,quad_real) - yv4
b3 = real(z,quad_real) - zv4

! compute the barycentric coordinates of the point

! det = A11*(A22*A33-A32*A23) - &
!       A21*(A12*A33-A32*A13) + &
!       A31*(A12*A23-A22*A13)

sump = 0.0_quad_real; summ = 0.0_quad_real
temp =  A11*A22*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -A11*A32*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -A21*A12*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  A21*A32*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  A31*A12*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -A31*A22*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
det = sump + summ

if (real(det,ce_real) == 0.0_ce_real) then
   ierr = 1
   return
endif

! compute the coordinates

! zeta(1) = (b1*(A22*A33-A32*A23) - &
!            b2*(A12*A33-A32*A13) + &
!            b3*(A12*A23-A22*A13))/det

sump = 0.0_quad_real; summ = 0.0_quad_real
temp =  b1*A22*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b1*A32*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b2*A12*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b2*A32*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b3*A12*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b3*A22*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
zeta1 = (sump + summ)/det
zeta(1) = zeta1

zeta(2) = (A11*(b2*A33-b3*A23) - &
           A21*(b1*A33-b3*A13) + &
           A31*(b1*A23-b2*A13))/det

sump = 0.0_quad_real; summ = 0.0_quad_real
temp =  b2*A11*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b3*A11*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b1*A21*A33
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b3*A21*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b1*A31*A23
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b2*A31*A13
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
zeta2 = (sump + summ)/det
zeta(2) = zeta2

! zeta(3) = (A11*(A22*b3-A32*b2) - &
!            A21*(A12*b3-A32*b1) + &
!            A31*(A12*b2-A22*b1))/det

sump = 0.0_quad_real; summ = 0.0_quad_real
temp =  b3*A11*A22
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b2*A11*A32
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b3*A21*A12
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b1*A21*A32
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp =  b2*A31*A12
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
temp = -b1*A31*A22
if (temp > 0.0_quad_real) then
   sump = sump + temp
else
   summ = summ + temp
endif
zeta3 = (sump + summ)/det
zeta(3) = zeta3

zeta(4) = 1.0_quad_real - zeta1 - zeta2 - zeta3

end subroutine accurate_barycentric_tet

end module containing_element
